@@ -1,7 +1,18 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
-# Copyright (C) 2006 Lukáš Lalinský
+#
+# Copyright (C) 2006-2008, 2011 Lukáš Lalinský
+# Copyright (C) 2009-2011, 2015, 2018-2020 Philipp Wolfer
+# Copyright (C) 2011 Johannes Weißl
+# Copyright (C) 2011-2014 Wieland Hoffmann
+# Copyright (C) 2012-2013 Michael Wiencek
+# Copyright (C) 2013 Calvin Walton
+# Copyright (C) 2013 Frederik “Freso” S. Olesen
+# Copyright (C) 2013-2014, 2018-2019 Laurent Monin
+# Copyright (C) 2014-2015 Sophist-UK
+# Copyright (C) 2016-2018 Sambhav Kothari
+# Copyright (C) 2019 Reinaldo Antonio Camargo Rauch
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -17,15 +28,16 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
+import re
+
 from mutagen.mp4 import (
     MP4,
     MP4Cover,
 )
 
-from picard import (
-    config,
-    log,
-)
+from picard import log
+from picard.config import get_config
 from picard.coverart.image import (
     CoverArtImageError,
     TagCoverArtImage,
@@ -36,6 +48,21 @@ from picard.metadata import Metadata
 from picard.util import encode_filename
 
 
+def _add_text_values_to_metadata(metadata, name, values):
+    for value in values:
+        metadata.add(name, value.decode("utf-8", "replace").strip("\x00"))
+
+
+_VALID_KEY_CHARS = re.compile('^[\x00-\xff]+$')
+
+
+def _is_valid_key(key):
+    """
+    Return true if a string is a valid name for a custom tag.
+    """
+    return bool(_VALID_KEY_CHARS.match(key))
+
+
 class MP4File(File):
     EXTENSIONS = [".m4a", ".m4b", ".m4p", ".m4v", ".mp4"]
     NAME = "MPEG-4 Audio"
@@ -51,8 +78,9 @@ class MP4File(File):
         "\xa9day": "date",
         "\xa9gen": "genre",
         "\xa9lyr": "lyrics",
-        "\xa9cmt": "comment:",
+        "\xa9cmt": "comment",
         "\xa9too": "encodedby",
+        "\xa9dir": "director",
         "cprt": "copyright",
         "soal": "albumsort",
         "soaa": "albumartistsort",
@@ -125,7 +153,8 @@ class MP4File(File):
     }
     __r_freeform_tags = dict([(v, k) for k, v in __freeform_tags.items()])
 
-    # Tags to load case insensitive
+    # Tags to load case insensitive. Case is preserved, but the specified case
+    # is written if it is unset.
     __r_freeform_tags_ci = {
         "replaygain_album_gain": "----:com.apple.iTunes:REPLAYGAIN_ALBUM_GAIN",
         "replaygain_album_peak": "----:com.apple.iTunes:REPLAYGAIN_ALBUM_PEAK",
@@ -159,28 +188,31 @@ class MP4File(File):
                 metadata.add(self.__bool_tags[name], values and '1' or '0')
             elif name in self.__int_tags:
                 for value in values:
-                    metadata.add(self.__int_tags[name], str(value))
+                    metadata.add(self.__int_tags[name], value)
             elif name in self.__freeform_tags:
-                for value in values:
-                    value = value.decode("utf-8", "replace").strip("\x00")
-                    metadata.add(self.__freeform_tags[name], value)
+                tag_name = self.__freeform_tags[name]
+                _add_text_values_to_metadata(metadata, tag_name, values)
             elif name_lower in self.__freeform_tags_ci:
-                for value in values:
-                    value = value.decode("utf-8", "replace").strip("\x00")
-                    tag_name = self.__freeform_tags_ci[name_lower]
-                    metadata.add(tag_name, value)
-                    self.__casemap[tag_name] = name
+                tag_name = self.__freeform_tags_ci[name_lower]
+                self.__casemap[tag_name] = name
+                _add_text_values_to_metadata(metadata, tag_name, values)
             elif name == "----:com.apple.iTunes:fingerprint":
                 for value in values:
                     value = value.decode("utf-8", "replace").strip("\x00")
                     if value.startswith("MusicMagic Fingerprint"):
                         metadata.add("musicip_fingerprint", value[22:])
             elif name == "trkn":
-                metadata["tracknumber"] = str(values[0][0])
-                metadata["totaltracks"] = str(values[0][1])
+                try:
+                    metadata["tracknumber"] = values[0][0]
+                    metadata["totaltracks"] = values[0][1]
+                except IndexError:
+                    log.debug('trkn is invalid, ignoring')
             elif name == "disk":
-                metadata["discnumber"] = str(values[0][0])
-                metadata["totaldiscs"] = str(values[0][1])
+                try:
+                    metadata["discnumber"] = values[0][0]
+                    metadata["totaldiscs"] = values[0][1]
+                except IndexError:
+                    log.debug('disk is invalid, ignoring')
             elif name == "covr":
                 for value in values:
                     if value.imageformat not in (value.FORMAT_JPEG,
@@ -197,12 +229,24 @@ class MP4File(File):
                                   (filename, e))
                     else:
                         metadata.images.append(coverartimage)
+            # Read other freeform tags always case insensitive
+            elif name.startswith('----:com.apple.iTunes:'):
+                tag_name = name_lower[22:]
+                self.__casemap[tag_name] = name[22:]
+                if (name not in self.__r_text_tags
+                    and name not in self.__r_bool_tags
+                    and name not in self.__r_int_tags
+                    and name not in self.__r_freeform_tags
+                    and name_lower not in self.__r_freeform_tags_ci
+                    and name not in self.__other_supported_tags):
+                    _add_text_values_to_metadata(metadata, tag_name, values)
 
         self._info(metadata, file)
         return metadata
 
     def _save(self, filename, metadata):
         log.debug("Saving file %r", filename)
+        config = get_config()
         file = MP4(encode_filename(self.filename))
         if file.tags is None:
             file.add_tags()
@@ -214,6 +258,8 @@ class MP4File(File):
         for name, values in metadata.rawitems():
             if name.startswith('lyrics:'):
                 name = 'lyrics'
+            if name == 'comment:':
+                name = 'comment'
             if name in self.__r_text_tags:
                 tags[self.__r_text_tags[name]] = values
             elif name in self.__r_bool_tags:
@@ -236,20 +282,38 @@ class MP4File(File):
                 tags[name] = values
             elif name == "musicip_fingerprint":
                 tags["----:com.apple.iTunes:fingerprint"] = [b"MusicMagic Fingerprint%s" % v.encode('ascii') for v in values]
+            elif self.supports_tag(name) and name not in self.__other_supported_tags:
+                values = [v.encode("utf-8") for v in values]
+                name = self.__casemap.get(name, name)
+                tags['----:com.apple.iTunes:' + name] = values
 
         if "tracknumber" in metadata:
-            if "totaltracks" in metadata:
-                tags["trkn"] = [(int(metadata["tracknumber"]),
-                                 int(metadata["totaltracks"]))]
+            try:
+                tracknumber = int(metadata["tracknumber"])
+            except ValueError:
+                pass
             else:
-                tags["trkn"] = [(int(metadata["tracknumber"]), 0)]
+                totaltracks = 0
+                if "totaltracks" in metadata:
+                    try:
+                        totaltracks = int(metadata["totaltracks"])
+                    except ValueError:
+                        pass
+                tags["trkn"] = [(tracknumber, totaltracks)]
 
         if "discnumber" in metadata:
-            if "totaldiscs" in metadata:
-                tags["disk"] = [(int(metadata["discnumber"]),
-                                 int(metadata["totaldiscs"]))]
+            try:
+                discnumber = int(metadata["discnumber"])
+            except ValueError:
+                pass
             else:
-                tags["disk"] = [(int(metadata["discnumber"]), 0)]
+                totaldiscs = 0
+                if "totaldiscs" in metadata:
+                    try:
+                        totaldiscs = int(metadata["totaldiscs"])
+                    except ValueError:
+                        pass
+                tags["disk"] = [(discnumber, totaldiscs)]
 
         covr = []
         for image in metadata.images.to_be_saved_to_tags():
@@ -274,18 +338,18 @@ class MP4File(File):
 
     @classmethod
     def supports_tag(cls, name):
-        return (name in cls.__r_text_tags
-                or name in cls.__r_bool_tags
-                or name in cls.__r_freeform_tags
-                or name in cls.__r_freeform_tags_ci
-                or name in cls.__r_int_tags
-                or name in cls.__other_supported_tags
-                or name.startswith('lyrics:')
-                or name in ('~length', 'musicip_fingerprint'))
+        unsupported_tags = ['r128_album_gain', 'r128_track_gain']
+        return ((name
+                 and not name.startswith("~")
+                 and name not in unsupported_tags
+                 and not (name.startswith('comment:') and len(name) > 9)
+                 and not name.startswith('performer:')
+                 and _is_valid_key(name))
+                or name in ('~length'))
 
     def _get_tag_name(self, name):
         if name.startswith('lyrics:'):
-            return 'lyrics'
+            name = 'lyrics'
         if name in self.__r_text_tags:
             return self.__r_text_tags[name]
         elif name in self.__r_bool_tags:
@@ -302,8 +366,9 @@ class MP4File(File):
             return "trkn"
         elif name in ("discnumber", "totaldiscs"):
             return "disk"
-        else:
-            return None
+        elif self.supports_tag(name) and name not in self.__other_supported_tags:
+            name = self.__casemap.get(name, name)
+            return '----:com.apple.iTunes:' + name
 
     def _info(self, metadata, file):
         super()._info(metadata, file)
@@ -312,5 +377,5 @@ class MP4File(File):
         filename = file.filename
         if isinstance(filename, bytes):
             filename = filename.decode()
-        if filename.lower().endswith(".m4v"):
+        if filename.lower().endswith(".m4v") or (file.tags and 'hdvd' in file.tags):
             metadata['~video'] = '1'
